/* ZMBV Commandline Encoder program
 * Copyright (C) 2013 Joel Yliluoma (http://iki.fi/bisqwit/)
 */

#include <string>
#include <vector>
#include <map>

#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <unistd.h>
#include <sys/stat.h>
#include <stdarg.h>
#include <vector>
#include <cerrno>

#include <assert.h>

#include "nesvideos-piece.hh"
#include "quantize.hh"
#include "lanczos.hh"
#include "rgbtorgb.hh"
#include <iostream>

#include "temporal_antialias.hh"

#include <getopt.h>

namespace
{
    /* Input stream specs */
    std::string VideoInputFileName="/dev/null";
    unsigned    IN_BPP=16, OUT_BPP=16;
    unsigned    ORIGWIDTH=320, ORIGHEIGHT=200;
    unsigned    WIDTH=320, HEIGHT=200;
    bool SwapRedBlue = true;
    // FPS, represented as a ratio of two integers:
    unsigned long long FPSHI=1,    FPSLO=1;
    unsigned long long FPSNEWHI=1, FPSNEWLO=1;
    enum fpsscaling_type { no_dont, yes_do, automatic } DoFPSscaling = automatic;
    /**** AUDIO ****/
    unsigned RATE=11025, CHANNELS=1, BITS=16;
    std::string AudioInputFileName="/dev/null";
    /**** LOGO ****/
    unsigned DELAY_BEGIN=0, FIRST_FRAME=0, LAST_FRAME=0x7FFFFFFFu;
    // Input will be delayed by this amount
    // (to allow for logo to display without blocking input):
    std::string EncodingRecipe="cat >/dev/null";
}

struct OverlayFun
{
    static void Overlay24With24
        (unsigned char* target, const unsigned char* source, int howmuch)
    {
        for(unsigned c=0; c<3; ++c)
            target[c] += ((int)(source[c] - target[c])) * howmuch / 255;
    }

    static bool Prefer32bitTempFormat()
    {
        return (OUT_BPP == 32 || OUT_BPP == 12 || OUT_BPP == 17
                            /*|| OUT_BPP == 15 || OUT_BPP == 16 // - no advantage*/
               )
            && (IN_BPP == 32 || IN_BPP == 12 || IN_BPP == 17
                             || IN_BPP == 15 || IN_BPP == 16
               );
    }

    static void ConvertInputVideoBPP
        (const char* VidBuf, std::vector<unsigned char>& logodata,
         int from_bpp)
    {
        /* RGB24 or RGB32 filled with black. */
        logodata.resize(WIDTH*HEIGHT * (Prefer32bitTempFormat() ? 4 : 3));
        if(from_bpp == 16)
        {
            if(Prefer32bitTempFormat())
                Convert16To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, !SwapRedBlue);
            else
                Convert16To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, !SwapRedBlue);
        }
        else if(from_bpp == 15)
        {
            if(Prefer32bitTempFormat())
                Convert15To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, SwapRedBlue);
            else
                Convert15To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, SwapRedBlue);
        }
        else if(from_bpp == 17)
        {
            if(Prefer32bitTempFormat())
                Convert_YUY2To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT*2, WIDTH, SwapRedBlue);
            else
                Convert_YUY2To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT*2, WIDTH, SwapRedBlue);
        }
        else if(from_bpp == 12)
        {
            if(Prefer32bitTempFormat())
                Convert_I420To32Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH, SwapRedBlue);
            else
                Convert_I420To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT, WIDTH, SwapRedBlue);
        }
        else if(from_bpp == 32)
        {
            if(Prefer32bitTempFormat())
                std::memcpy(&logodata[0], VidBuf, WIDTH*HEIGHT*4);
            else
                Convert32To24Frame(VidBuf, &logodata[0], WIDTH*HEIGHT);
        }
        else if(from_bpp == 24)
            std::memcpy(&logodata[0], VidBuf, WIDTH*HEIGHT*3);
        else
        {
            std::fprintf(stderr, "hudmaker: I do not know how to handle input bpp %u\n", from_bpp);
            std::abort();
        }
    }

    static void ConvertInputVideoBPP
        (const char* VidBuf, std::vector<unsigned char>& logodata)
    {
        ConvertInputVideoBPP(VidBuf, logodata, IN_BPP);
    }

    static void ConvertOutputVideoBPP
        (char* VidBuf, const std::vector<unsigned char>& logodata,
         int out_bpp)
    {
        if(out_bpp == 16)
        {
            if(Prefer32bitTempFormat())
                Convert32To16Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To16Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(out_bpp == 15)
        {
            if(Prefer32bitTempFormat())
                Convert32To15Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To15Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(out_bpp == 17)
        {
            if(Prefer32bitTempFormat())
                Convert32To_YUY2Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To_YUY2Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(out_bpp == 12)
        {
            if(Prefer32bitTempFormat())
                Convert32To_I420Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
            else
                Convert24To_I420Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH);
        }
        else if(out_bpp == 8)
            Convert24To8Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT, WIDTH, SwapRedBlue);
        else if(out_bpp == 24)
        {
            std::memcpy(VidBuf, &logodata[0], logodata.size());
        }
        else if(out_bpp == 32/*)
        {
            if(*/&& Prefer32bitTempFormat())
                std::memcpy(VidBuf, &logodata[0], logodata.size());
            /*else
                Convert24To32Frame(&logodata[0], (unsigned char*) VidBuf, WIDTH*HEIGHT);
        }*/
        else
        {
            std::fprintf(stderr, "hudmaker: I do not know how to handle output bpp %u\n", out_bpp);
            std::abort();
        }
    }

    static void ConvertOutputVideoBPP
        (char* VidBuf, const std::vector<unsigned char>& logodata)
    {
        ConvertOutputVideoBPP(VidBuf, logodata, OUT_BPP);
    }
};

class FrameToInterpolate
{
public:
    char* VidBuf;
    FrameToInterpolate(char* f) : VidBuf(f) { }
    FrameToInterpolate(): VidBuf(0) { }
    void donewith()
    {
        std::free(VidBuf);
    }
};

//const double Gamma = 2.0, Ungamma = 1.0 / Gamma;
//std::vector<Triplet<double> > CurrentFrameBuffer;
std::vector<Triplet<unsigned> > CurrentFrameBuffer;
void InterpolateFrames(FrameToInterpolate& f, int weight)
{
    std::vector<unsigned char> frame;
    OverlayFun::ConvertInputVideoBPP(f.VidBuf, frame);
    unsigned n = OverlayFun::Prefer32bitTempFormat() ? 4 : 3;
    #pragma omp parallel for
    for(unsigned y=0; y<HEIGHT; ++y)
    for(unsigned p=y*WIDTH, o=p*n, x=0; x<WIDTH; ++x, ++p,o+=n)
        for(int a=0; a<3; ++a)
            CurrentFrameBuffer[p][a] += frame[o+a] * weight;
}

static void fread2(void* target, size_t a, size_t b, FILE* fp)
{
    char* t = (char*) target;
    size_t nbytes = a * b;

    while(nbytes > 0)
    {
        int r = std::fread(t, 1, nbytes, fp);
        if(r < 0 && errno == EINTR) continue;
        if(r < 0 && errno == EAGAIN) continue;
        if(r <= 0) break;
        nbytes -= r;
        t      += r;
    }
}

static void fseek2_rel(FILE* fp, long n_bytes)
{
    int r = fseek(fp, n_bytes, SEEK_CUR);
    if(r < 0)
    {
        char Buffer[4096];
        while(n_bytes > 0)
        {
            long take = 4096;
            if(take > n_bytes) take = n_bytes;
            fread2(Buffer, 1, take, fp);
            n_bytes -= take;
        }
    }
}

static void MakeFraction(
    long double orig,
    unsigned long long& hi, unsigned long long& lo,
    int maxdepth = 25)
{
    if(orig == 0.0) { hi=lo=0; return; }

    long long cf[25];
    for(int depth=1; depth<maxdepth; ++depth)
    {
        long long u,v;
        long double error, a=orig;
        int i, last;

        for(i = 0; i < depth; ++i)
        {
            cf[i] = (long long)a;
            if(cf[i] - (long long)1 > cf[i])break;
            a = 1.0 / (a - cf[i]);
        }

        for(last=i-1; i < depth; ++i)
            cf[i] = 0;

        u = cf[last];
        v = 1;
        for(i = last-1; i >= 0; --i)
        {
            long long w = cf[i] * u + v;
            v = u;
            u = w;
        }

        error = (orig - (u / (long double)v)) / orig;

        hi=u; lo=v;
        if(error < 1e-8 && error > -1e-8)break;
    }
}

static void ReadArgs(int argc, char** argv)
{
    VideoInputFileName = "/dev/stdin";
    AudioInputFileName = "/dev/null";
    std::string OutputFileName = "/dev/stdout";
    WIDTH=HEIGHT=OUT_BPP=0;
    double in_fps=0.0, out_fps=0.0;
    ORIGWIDTH=ORIGHEIGHT=0;

    for(;;)
    {
        int option_index = 0;
        static struct option long_options[] =
        {
            {"output",  1, 0, 'o'},
            {"help",    0, 0, 'h'},
            {"input",   1, 0, 'i'},
            {"audio",   1, 0, 'I'},
            {"width",   1, 0, 'x'},
            {"height",  1, 0, 'y'},
            {"bpp",     1, 0, 'b'},
            {"tgtwidth",  1, 0, 'X'},
            {"tgtheight", 1, 0, 'Y'},
            {"tgtbpp",    1, 0, 'B'},
            {"fps",       1, 0, 401},
            {"tgtfps",    1, 0, 402},
            {"swapredblue", 0,0,403},
            {"rate",      1, 0, 'r'},
            {"channels",  1, 0, 'c'},
            {"bits",      1, 0, 's'},
            {"quant",     1, 0, 'q'},
            {0,0,0,0}
        };
        int c = getopt_long(argc, argv, "o:hi:I:x:y:b:X:Y:B:r:c:s:q:",
                            long_options, &option_index);
        if(c == -1) break;
        switch(c)
        {
            case 'V':
                std::printf("%s\n", VERSION);
                std::exit(0);

            case 'h':
                std::printf(
                    "Options:\n"
                    " --help, -h           This help\n"
                    " --output, -o <file>  *Output into given file\n"
                    " --input, -i <file>   Input raw video from given file (default: /dev/stdin)\n"
                    " --audio, -I <file>   Input raw audio from given file (default: /dev/null)\n"
                    " --width, -x <n>     *Video width is n pixels\n"
                    " --height, -y <n>    *Video height is n pixels\n"
                    " --bpp, -b <n>       *Video bpp is n. List of supported bpp are below. (default: 16)\n"
                    "                          12 = I420\n"
                    "                          17 = YUY2\n"
                    "                          15 = bgr15\n"
                    "                          16 = bgr16\n"
                    "                          24 = bgr24\n"
                    "                          32 = bgra / bgr32\n"
                    " --swapredblue        Swap red & blue in some BPP conversions.\n"
                    " --tgtwidth, -X <n>   Video is to be scaled into n pixels width (default: source width)\n"
                    " --tgtheight, -Y <n>  Video is to be scaled into n pixels height (default: source height)\n"
                    " --tgtbpp, -B <n>     Video is to be converted into bpp n (default: source bpp)\n"
                    " --fps <n>           *Specify source video framerate (default: 1)\n"
                    " --tgtfps <n>         Specify target video framerate (default: source framerate)\n"
                    " --rate, -r <n>       Audio on a channel is n samples per second (default: 11025)\n"
                    " --channels, -c <n>   Audio has n channels (default: 1)\n"
                    " --bits, -s <n>       Audio samples are n bits wide (default: 16)\n"
                    " --quant, -q <n>      Quantization control\n"
                    "      -q 0   Never quantize. Always use RGB formats for ZMBV.\n"
                    "      -q 1   Always quantize, and use a 5*7*7 regular RGB palette.\n"
                    "      -q 2   Always quantize, and use a hardcoded RGB palette that contains\n"
                    "             colors compatible with both xterm-256color and VGA, as well as\n"
                    "             some tones of purple.\n"
                    "      -q 3   When quantizing, use fast regular dithering\n"
                    "             (works only with the 5*7*7 regular RGB palette)\n"
                    "      -q 4   When quantizing, use positional dithering with gamma correction\n"
                    "             (works with all palettes)\n"
                    "      -q 5   Never quantize; use RGB. However, if the input video\n"
                    "             contains 256 colors or less, paletted format is saved.\n"
                );
                std::exit(0);

            case 'o':
                OutputFileName = optarg;
                break;

            case 'i': VideoInputFileName = optarg; break;
            case 'I': AudioInputFileName = optarg; break;
            case 'x': ORIGWIDTH  = std::strtol(optarg, 0, 10); break;
            case 'y': ORIGHEIGHT = std::strtol(optarg, 0, 10); break;
            case 'b': IN_BPP  = std::strtol(optarg, 0, 10); break;
            case 'X': WIDTH   = std::strtol(optarg, 0, 10); break;
            case 'Y': HEIGHT  = std::strtol(optarg, 0, 10); break;
            case 'B': OUT_BPP = std::strtol(optarg, 0, 10); break;
            case 401: in_fps  = std::strtod(optarg, 0); break;
            case 402: out_fps = std::strtod(optarg, 0); break;
            case 403: SwapRedBlue = false; break;
            case 'r': RATE     = std::strtol(optarg, 0, 10); break;
            case 'c': CHANNELS = std::strtol(optarg, 0, 10); break;
            case 's': BITS     = std::strtol(optarg, 0, 10); break;
            case 'q':
                switch(std::strtol(optarg, 0, 10))
                {
                    case 0:
                        ZMBV_NEVER_QUANTIZE = true;
                        ZMBV_ALLOWED_TO_SAVE_PALETTED_WITHOUT_CHANGES = false;
                        break;
                    case 1:
                        ZMBV_NEVER_QUANTIZE = false;
                        ZMBV_QUANTIZE_CUSTOM_PALETTE = false;
                        ZMBV_QUANTIZE_ADVANCED_DITHERING = false;
                        break;
                    case 2:
                        ZMBV_NEVER_QUANTIZE = false;
                        ZMBV_QUANTIZE_CUSTOM_PALETTE = true;
                        break;
                    case 3:
                        ZMBV_QUANTIZE_ADVANCED_DITHERING = false;
                        break;
                    case 4:
                        ZMBV_QUANTIZE_ADVANCED_DITHERING = true;
                        break;
                    case 5:
                        ZMBV_NEVER_QUANTIZE = true;
                        ZMBV_ALLOWED_TO_SAVE_PALETTED_WITHOUT_CHANGES = true;
                        break;
                    default:
                        printf("Illegal parameter: -q%s\n", optarg);
                        break;
                }
                break;
        }
    }
    if(argc != optind)
    {
        std::printf("Invalid parameters. Use the --help option for list of valid parameters.\n");
        std::exit(1);
    }
    if(!ORIGWIDTH || !ORIGHEIGHT)
    {
        std::printf("At least please tell me the input video geometry. Use the --help option.\n");
        std::exit(1);
    }

    if(!WIDTH)   WIDTH = ORIGWIDTH;
    if(!HEIGHT)  HEIGHT = ORIGHEIGHT;
    if(!OUT_BPP) OUT_BPP = IN_BPP;
    if(out_fps == 0.0) out_fps = in_fps;

    if(OUT_BPP == 12 || OUT_BPP == 17)
    {
        std::printf("In ZMBV version, I420 and YUY2 formats cannot be used. Converting to RGB32.\n");
        OUT_BPP = 32;
    }
    if(OUT_BPP == 8)
    {
        std::printf("In ZMBV version, BPP 8 does not work quite right. Converting to 32bpp instead.\n");
        OUT_BPP = 32;
    }

    MakeFraction(in_fps,  FPSHI,   FPSLO);
    MakeFraction(out_fps, FPSNEWHI,FPSNEWLO);

    EncodingRecipe = "MKV:" + OutputFileName;
}

int main(int argc, char**argv)
{
    ReadArgs(argc, argv);

    LoggingEnabled = 2;
    NESVideoSetVideoCmd(EncodingRecipe.c_str());
    NESVideoOpenNow(); // Ensure it is opened before OpenMP is used

    FILE* ai = std::fopen(AudioInputFileName.c_str(), "rb");
    FILE* vi = std::fopen(VideoInputFileName.c_str(), "rb");
    if(!ai && RATE) { std::perror(AudioInputFileName.c_str()); }
    if(!vi) { std::perror(VideoInputFileName.c_str()); }
    if(!ai || !vi) return -1;

    const unsigned fps_scaled = FPSNEWHI * 0x1000000ull / FPSNEWLO;
    unsigned long audio_samples = 0;
    unsigned remaining_delay = DELAY_BEGIN;

    long framewait = FIRST_FRAME;

    // Disable buffering for the streams to avoid stdio reading a piece
    // of the file on each successive seek. Also, we know exactly how
    // much we want to read each time, so stdio buffering is not helpful.
    if(ai && RATE) std::setbuf(ai, NULL);
    if(vi) std::setbuf(vi, NULL);

    if(FPSLO == 0)    std::fprintf(stderr, "ERROR: FPSLO = 0, THUS INPUT FPS = INFINITE\n");
    if(FPSNEWLO == 0) std::fprintf(stderr, "ERROR: FPSNEWLO = 0, THUS OUTPUT FPS = INFINITE\n");

    double NewFPS = (double)FPSNEWHI / (double)FPSNEWLO;
    double OldFPS = (double)FPSHI    / (double)FPSLO;

    bool Need_FrameMerging = NewFPS < OldFPS ||
        (0.5 - std::abs(0.5 - std::fmod(NewFPS/OldFPS, 1.0))) > 1e-2;
    if(DoFPSscaling == no_dont) Need_FrameMerging = false;
    if(DoFPSscaling == yes_do) Need_FrameMerging = true;

    if(IN_BPP != OUT_BPP) Need_FrameMerging = true;

    TemporalAntialias<FrameToInterpolate>
        FrameInterpolater( (double)FPSNEWLO / (double)FPSNEWHI );

    for(long framecounter=0; framecounter<LAST_FRAME; ++framecounter)
    {
        const unsigned bytes_per_60_in_pixels = (IN_BPP == 15 ? 16 : IN_BPP == 17 ? 16 : IN_BPP) * 60 / 8;
        const unsigned bytes_per_60_out_pixels =(OUT_BPP== 15 ? 16 :OUT_BPP == 17 ? 16 :OUT_BPP) * 60 / 8;
        const unsigned long in_image_size = WIDTH * HEIGHT * bytes_per_60_in_pixels / 60;
        const unsigned long out_image_size= WIDTH * HEIGHT * bytes_per_60_out_pixels / 60;
        char* VidBuf = (char*)
            std::malloc( in_image_size > out_image_size ? in_image_size : out_image_size );

        const unsigned bytes_per_60_samples = CHANNELS * 60 * BITS / 8;
        audio_samples += RATE * FPSLO;
        const unsigned take_audio_samples = audio_samples / FPSHI;
        const unsigned take_audio_bytes = bytes_per_60_samples/60*take_audio_samples;
        char* AudBuf = (char*)
            std::malloc( RATE * bytes_per_60_samples / 60 ); // Assume FPS is always >= 1.000

        audio_samples -= take_audio_samples * FPSHI;

        if(remaining_delay > 0)
        {
            std::memset(VidBuf, 0, in_image_size);
            std::memset(AudBuf, 0, take_audio_bytes);
            --remaining_delay;
        }
        else
        {
            if( (IN_BPP == 12 || IN_BPP == 17)
            && (ORIGWIDTH != WIDTH || ORIGHEIGHT != HEIGHT) )
            {
                std::fprintf(stderr, "hudmaker: Cannot scale chroma-supersampled images\n");
                std::abort();
            }
            const unsigned by = bytes_per_60_in_pixels / 60;
            const unsigned original_video_frame_size =
                ORIGWIDTH * ORIGHEIGHT * bytes_per_60_in_pixels / 60;

            if(framewait > 0)
            {
                if(ai && RATE) fseek2_rel(ai, take_audio_bytes);
                if(vi) fseek2_rel(vi, original_video_frame_size);
            }
            else
            {
                if(ai && RATE)
                {
                    fread2(AudBuf, bytes_per_60_samples/60, take_audio_samples, ai);
                }
                if(ORIGWIDTH == WIDTH && ORIGHEIGHT == HEIGHT)
                {
                    fread2(VidBuf, 1, original_video_frame_size, vi);
                }
                else
                {
                    static std::vector<char> TempBuf( original_video_frame_size );
                    /* This buffer is "static" because otherwise, at
                     * high resolutions we can easily run out of stack space.
                     */
                    fread2(&TempBuf[0], 1, original_video_frame_size, vi);
                    /* Point-scale the original image to the resulting image's size */

                    /*if((IN_BPP == 24
                     || IN_BPP == 32)
                    && WIDTH > ORIGWIDTH*2 && HEIGHT > ORIGHEIGHT*2)
                    {
                        std::vector<Triplet<unsigned char> > input_frame
                            ( ORIGWIDTH*2 * ORIGHEIGHT*2 );
                        std::vector<Triplet<LanczosFloat> >  scaled_frame;

                        //#pragma omp parallel for schedule(static) num_threads(3)
                        for(unsigned y=0; y<ORIGHEIGHT; ++y)
                        {
                            const unsigned char* src =
                                (const unsigned char*) &TempBuf[0]
                                + y*ORIGWIDTH*by;
                            unsigned p = (y*2) * (ORIGWIDTH*2);
                            for(unsigned x=0; x<ORIGWIDTH; ++x, p+=2, src+=by)
                            {
                                input_frame[p+1 +ORIGWIDTH*2] =
                                input_frame[p   +ORIGWIDTH*2] =
                                input_frame[p+1] =
                                input_frame[p] =
                                    Triplet<unsigned char>
                                        (src[0], src[1], src[2], (unsigned char)0);
                            }
                        }
                        LanczosScale(ORIGWIDTH*2, ORIGHEIGHT*2, WIDTH,HEIGHT,
                            input_frame, scaled_frame);
                        #pragma omp parallel for schedule(static) num_threads(3)
                        for(unsigned y=0; y<HEIGHT; ++y)
                        {
                            char* tgt = VidBuf + y*WIDTH*by;
                            unsigned p = y*WIDTH;
                            for(unsigned x=0; x<WIDTH; ++x,++p)
                            {
                                const int o16 = (x+4*y)%16;
                                int r = Quantize4x4<0xFF> (o16, scaled_frame[p][0]); if(r<0)r=0; if(r>255)r=255;
                                int g = Quantize4x4<0xFF> (o16, scaled_frame[p][1]); if(g<0)r=0; if(g>255)r=255;
                                int b = Quantize4x4<0xFF> (o16, scaled_frame[p][2]); if(b<0)r=0; if(b>255)r=255;
                                tgt[0] = r;
                                tgt[1] = g;
                                tgt[2] = b;
                                tgt += by;
                            }
                        }
                    }
                    else
                    */
                    {
                        #if 0
                        std::memset(VidBuf, 0, out_image_size);
                        #endif

                        #pragma omp parallel for schedule(static)
                        for(unsigned y=0; y< ORIGHEIGHT; ++y)
                        {
                            const unsigned tgt_y = y * HEIGHT / ORIGHEIGHT;
                            const unsigned yscale = (y+1)*HEIGHT / ORIGHEIGHT - tgt_y;

                            char* src = &TempBuf[0] + y*ORIGWIDTH*by;
                            char* tgt = VidBuf      + tgt_y*WIDTH*by;

                            unsigned xcarry = 0;
                            for(unsigned x=0; x< ORIGWIDTH; ++x)
                            {
                            #if 1
                                xcarry += WIDTH;
                                const unsigned xscale = xcarry / ORIGWIDTH;
                                xcarry %= ORIGWIDTH;
                            #else
                                const unsigned xscale = 2;
                            #endif

                                unsigned xdiff = xscale*by;
                                for(unsigned ys=0, ysp=yscale*WIDTH; ys<ysp; ys+=WIDTH)
                                {
                                    char* x_tgt     = &tgt[by*ys];
                                    char* x_tgt_end = x_tgt + xdiff;
                                    while(x_tgt < x_tgt_end)
                                        { std::memcpy(x_tgt, src, by);
                                          x_tgt += by; }
                                }
                                /*for(unsigned c=0; c<by; ++c, ++tgt)
                                {
                                    char byt = *src++;
                                    for(unsigned ys=0, ysp=yscale*WIDTH; ys<ysp; ys += WIDTH)
                                    for(unsigned xs=0; xs<xscale; ++xs)
                                        tgt[by * (xs + ys)] = byt;
                                }*/
                                tgt += by * (xscale/*-1*/);
                                src += by;
                            }
                        }
                    } // lanczos condition
                } // scale condition
            } // frameskip condition
        } // delay condition

        if(framewait > 0)
        {
            --framewait;
            std::free(VidBuf);
            std::free(AudBuf);
        }
        else
        {
            static double when = 0.0;

            // Use frame drop&dup method?
            if(!Need_FrameMerging)
            {
                static unsigned long long cache = 0;
                cache += FPSLO*FPSNEWHI;
                bool v_used = false;
                while(cache >= FPSNEWLO*FPSHI)
                {
                    char* vframe = VidBuf;
                    v_used = true;
                    cache -= FPSNEWLO*FPSHI;
                    if(cache >= FPSNEWLO*FPSHI) // Will this frame be duplicated?
                    {
                        // If so, make a copy of VidBuf so that the
                        // next call to AddTo() won't deallocate it
                        // when it still will be needed
                        VidBuf = (char*)std::malloc(out_image_size);
                        memcpy(VidBuf, vframe, out_image_size);
                    }
                    NESVideoLoggingVideoPtr(vframe, WIDTH, HEIGHT, fps_scaled, OUT_BPP);
                }
                if(!v_used) std::free(VidBuf);
            }
            else // Use temporal smoothing
            {
                unsigned n = OverlayFun::Prefer32bitTempFormat() ? 4 : 3;
                /*if(OUT_BPP != n*8)
                {
                    std::fprintf(stderr, "SORRY: OUT_BPP MUST BE %u\n", n*8);
                    std::exit(-1);
                }*/

                FrameToInterpolate f(VidBuf); // Transfers ownership
                FrameInterpolater.FeedAndEat(when, f);
                FrameInterpolater.SetMinim(0.02);
                //    when < 400 ? 0.02 : 0.1 );
                while(FrameInterpolater.Avail())
                {
                    // Initialize to zero
                    CurrentFrameBuffer.clear();
                    CurrentFrameBuffer.resize( WIDTH*HEIGHT );
                    FrameInterpolater.Get(InterpolateFrames//,
                        //when < 1.8
                                         );

                    char* vframe = (char*) std::calloc(WIDTH*HEIGHT, n);
                    // ^ deallocated by NESVideo
                    #pragma omp parallel for
                    for(unsigned y=0; y<HEIGHT; ++y)
                    for(unsigned p=y*WIDTH, o=p*n, x=0; x<WIDTH; ++x, ++p,o+=n)
                    {
                        //const int o16 = (x+4*y)%16;
                        for(unsigned c=0; c<3; ++c)
                        {
                            int v = CurrentFrameBuffer[p][c] / 4194304;
                            if(v > 255) v = 255;
                            vframe[o+c] = v;
                        }
                    }
                    NESVideoLoggingVideoPtr(vframe, WIDTH, HEIGHT, fps_scaled, OUT_BPP);
                }
            }

            when += (double)FPSLO / (double)FPSHI;

            NESVideoLoggingAudioPtr(AudBuf, RATE, BITS, CHANNELS, take_audio_samples);

            if(std::feof(vi) && std::feof(ai)) break;
        }
    }
    std::fclose(ai);
    std::fclose(vi);

    NESVideoFlush();
    return 0;
}
